/*  einsc -- a compiler for eins
 *    Semantic analyzer.
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Analyzer.java,v 1.2 2005/04/27 13:56:08 dragos Exp $
 */

package einsc;

import einsc.Tree.*;
import einsc.Type.ObjType;

/** This class implements the semantic analyzer of einsc.
 */
public class Analyzer implements Visitor {

    /** The outermost scope.
     */
    public final Scope globalScope;

    /** The current scope.
     */
    Scope scope;

    public Analyzer () {
        globalScope = new Scope();

        // enter a predefined Unit class into the top-level scope
        Scope unitFields = new Scope();
        Type unitClassType = new Type.ClassType(unitFields);
        Symbol unitClassSymbol = new Symbol(0, "Unit", unitClassType, Symbol.CLASS);
        globalScope.enter(unitClassSymbol);
        Type unitObjType = new Type.ObjType(unitClassSymbol);

        // enter a predefined List class into the top-level scope
        Scope listFields = new Scope();
        Type listClassType = new Type.ClassType(listFields);
        Symbol listClassSymbol = new Symbol(0, "List", listClassType, Symbol.CLASS);
        globalScope.enter(listClassSymbol);
        Type listObjType = new Type.ObjType(listClassSymbol);
        listFields.enter(new Symbol(0, "head", Type.INT_TYPE, Symbol.FIELD));
        listFields.enter(new Symbol(0, "tail", listObjType, Symbol.FIELD));

        // enter predefined functions into the top-level scope
        Symbol tmpSym = new Symbol(0,
                "printInt",
                new Type.FunType(new Type[]{Type.INT_TYPE}, unitObjType),
                Symbol.FUNC);
        tmpSym.label = "_printInt";
        globalScope.enter(tmpSym);
        
        tmpSym = new Symbol(0,
                "printChar",
                new Type.FunType(new Type[]{Type.INT_TYPE}, unitObjType),
                Symbol.FUNC);
        tmpSym.label = "_printChar";
        globalScope.enter(tmpSym);
        
        tmpSym = new Symbol(0,
                "readInt",
                new Type.FunType(new Type[]{}, Type.INT_TYPE),
                Symbol.FUNC);
        tmpSym.label = "_readInt";
        globalScope.enter(tmpSym);
        
        tmpSym = new Symbol(0,
                "readChar",
                new Type.FunType(new Type[]{}, Type.INT_TYPE),
                Symbol.FUNC); 
        tmpSym.label = "_readChar";
        globalScope.enter(tmpSym);

        scope = globalScope;
    }

    /** Analyze the tree with a given scope
     */
    Type analyze(Tree tree, Scope scope) {
        Scope backup = this.scope;
        this.scope = scope;
        Type type = analyze(tree);
        this.scope = backup;
        return type;
    }
    
    /** Analyze the tree in the current scope
     */
    public Type analyze(Tree tree) {
        tree.apply(this);
        return tree.type;
    }

    /** Analyze several trees with a given scope.
     */
    private Type[] analyze(Tree[] trees, Scope scope) {
        Type[] types = new Type[trees.length];
        for (int i = 0; i < trees.length; i++)
            types[i] = analyze(trees[i], scope);
        return types;
    }

    /** Analyze several trees in the current scope.
     */
    private Type[] analyze(Tree[] trees) {
        return analyze(trees, scope);
    }
    
    /** Generates an error if the expected type is not equal
     *  to the found type.
     */
    private void checkSameType(int pos, Type found, Type expected) {
        if (!found.isSameType(expected))
            error(pos, found, expected);
    }
    
//    private boolean checkTypes(Type[] types, Type expected) {
//        for (int i = 0; i < types.length; ++i)
//            if (!types[i].isSameType(expected)) 
//                return false;
//        return true;
//    }
    
    private boolean checkSubTypes(Type[] types, Type expected) {
        for (int i = 0; i < types.length; ++i)
            if (!types[i].isSubType(expected)) 
                return false;
        return true;
    }

    /** Generates an error if the expected type is not a super type
     *  of the found type.
     */
    private void checkSubType(int pos, Type found, Type expected){
        if (!found.isSubType(expected))
            error(pos, found, expected);
    }

    private void error(int pos, Type found, Type expected) {
        Report.error(pos, "expected type: " + expected + "\n" +
                          "actual type  : " + found);
    }
    
    //// IMPLEMENTATION OF THE VISITOR INTERFACE ////
    
    public void caseBad(Bad tree) {
        tree.type = Type.BAD_TYPE;
    }

    public void caseProgram(Program tree) {
        scope = new Scope(globalScope);
        analyze(tree.decls);
        analyze(tree.main);
        tree.type = Type.NO_TYPE;
    }

    public void caseClassDef(ClassDef tree) {
        Scope fields = new Scope();
        Type classType = new Type.ClassType(fields);
        Symbol classSym = new Symbol(tree.pos, tree.name, classType, Symbol.CLASS);
        scope.enter(classSym);

        Symbol[] fieldSyms = new Symbol[tree.fields.length];
        for (int i = 0; i < tree.fields.length; i++) {
            Formal field = (Formal) tree.fields[i];
            Symbol fieldSym = new Symbol(field.pos, field.name, analyze(field.vartype), Symbol.FIELD);
            field.sym = fieldSym;
            field.type = field.sym.type;
            fields.enter(fieldSym);
        }
        tree.sym = classSym;
        tree.type = Type.NO_TYPE;
    }

    public void caseFunDef(FunDef tree) {
        Tree[] paramtypes = new Tree[tree.params.length];
        for(int i = 0; i < tree.params.length; i++)
            paramtypes[i] = ((Formal) tree.params[i]).vartype;
        Type.FunType funtype = (Type.FunType) analyze(new Tree.FunType(0, paramtypes, tree.restype));
        tree.sym = new Symbol(tree.pos, tree.name, funtype, Symbol.FUNC);
        scope.enter(tree.sym);
        Scope local = new Scope(scope);
        analyze(tree.params, local);
        checkSubType(tree.pos, analyze(tree.body, local), funtype.resType);
        tree.type = Type.NO_TYPE;
    }

    public void caseFormal(Formal tree) {
        tree.sym = new Symbol(tree.pos, tree.name,
                              analyze(tree.vartype), Symbol.VAR);
        scope.enter(tree.sym);
        tree.type = tree.sym.type;
    }

    public void caseIntType(IntType tree) {
        tree.type = Type.INT_TYPE;
    }

    public void caseFunType(FunType tree) {
        tree.type = new Type.FunType(analyze(tree.paramtypes),
                                     analyze(tree.restype));
    }
    
    public void caseArrayType(ArrayType tree) {
        Type base = analyze(tree.baseType);
        tree.type = new Type.ArrayType(base, tree.dims);
    }

    public void caseTypeIdent(TypeIdent tree) {
        tree.sym = scope.lookup(tree.name);
        if (tree.sym == null || tree.sym.kind != Symbol.CLASS) {
            tree.type = Type.BAD_TYPE;
            Report.error(tree.pos, "the class " + tree.name + " is unknown.");
        } else
            tree.type = new Type.ObjType(tree.sym);
    }

    public void caseWhile(While tree) {
        // !!
        checkSubType(tree.pos, analyze(tree.cond), Type.INT_TYPE);
        Scope localScope = new Scope(scope);
        analyze(tree.stats, localScope);
        tree.type = Type.NO_TYPE;
    }

    public void caseVarDef(VarDef tree) {
        analyze(tree.init);
        analyze(tree.variable);
        checkSubType(tree.pos, tree.init.type, tree.variable.type);
        tree.type = Type.NO_TYPE;
    }
    
    public void caseAssign(Assign tree) {
//        tree.sym = scope.lookup(tree.name);
//        Type valueType = analyze(tree.expr);
//        if (tree.sym == null)
//            Report.error(tree.pos, "the variable " + tree.name + " is unknown.");
//        else if (tree.sym.kind != Symbol.VAR)
//            Report.error(tree.pos, tree.name + " does not refer to a variable.");
//        else
//            checkSubType(tree.pos, valueType, tree.sym.type);
//        tree.type = Type.NO_TYPE;
        Type leftType = analyze(tree.var);
        Type rightType = analyze(tree.expr);
        
        if (!leftType.isRefType()) {
            Report.error(tree.pos, "type of left side of the assignment is not assignable (" + leftType + ")");
        } else
            checkSubType(tree.pos, rightType, ((Type.RefType)leftType).baseType);
    }

    public void caseDo(Do tree) {
        analyze(tree.expr);
        tree.type = Type.NO_TYPE;
    }

    public void caseIf(If tree) {
        // !!
        checkSubType(tree.pos, analyze(tree.cond), Type.INT_TYPE);
        tree.type = analyze(tree.thenp).max(analyze(tree.elsep));
        if (tree.type == Type.BAD_TYPE)
            Report.error(tree.pos,
                "incompatible branches of 'if' statement\n" +
                "type of then-part: " + tree.thenp.type + "\n" +
                "type of else-part: " + tree.elsep.type);
    }

    public void caseUnop(Unop tree) {
        // !!
        checkSubType(tree.pos, analyze(tree.expr), Type.INT_TYPE);
        tree.type = Type.INT_TYPE;
    }

    public void caseBinop(Binop tree) {
        Type t1 = analyze(tree.left);
        Type t2 = analyze(tree.right);
        if (tree.op == Tree.EQ || tree.op == Tree.NE) {
            if (t1.max(t2) == Type.BAD_TYPE || (t1 instanceof Type.FunType))
                Report.error(tree.pos, "Incomparable operands (" + t1 + ", " + t2 + ")");
        }
        else {
            //!!
            checkSubType(tree.left.pos, t1, Type.INT_TYPE);
            checkSubType(tree.right.pos, t2, Type.INT_TYPE);
        }
        tree.type = Type.INT_TYPE;
    }

    public void caseIdent(Ident tree) {
        tree.sym = scope.lookup(tree.name);
        if (tree.sym == null) {
            Report.error(tree.pos, "the symbol " + tree.name + " is unknown.");
            tree.type = Type.BAD_TYPE;
        }
        else if (tree.sym.kind != Symbol.VAR && tree.sym.kind != Symbol.FUNC) {
            Report.error(tree.pos, "the symbol " + tree.name + " is neither a variable nor a function.");
            tree.type = Type.BAD_TYPE;
        }
        else
            // !!
            tree.type = tree.sym.kind == Symbol.VAR ? new Type.RefType(tree.sym.type) : tree.sym.type;
    }

    public void caseNullLit(NullLit tree) {
        tree.type = Type.NULL_TYPE;
    }
    
    public void caseIntLit(IntLit tree) {
        tree.type = Type.INT_TYPE;
    }

    public void caseApply(Apply tree) {
        Type funType = analyze(tree.fun);
        Type[] argTypes = analyze(tree.args);
        if (funType.isFunType()) {
            Type.FunType tpe = (Type.FunType) (funType.isRefType() ? ((Type.RefType)funType).baseType : funType);
            Type[] paramTypes = tpe.argTypes;
            if (argTypes.length != paramTypes.length)
                Report.error(tree.pos, "Bad number of arguments: " + tree);
            else {
                for(int i = 0; i < paramTypes.length; i++)
                    checkSubType(tree.pos, argTypes[i], paramTypes[i]);
            }
            tree.type = tpe.resType;
        }
        else {
            tree.type = Type.BAD_TYPE;
            if (funType != Type.BAD_TYPE)
                Report.error(tree.fun.pos, "Not a function: " + tree.fun);
        }
    }

    public void caseBlock(Block tree) {
        Scope local = new Scope(scope);
        analyze(tree.stats, local);
        tree.type = analyze(tree.main, local);
    }

    public void caseNew(New tree) {
        Symbol classSym = scope.lookup(tree.name);
        Type[] argTypes = analyze(tree.args);
        if (classSym == null) {
            tree.type = Type.BAD_TYPE;
            Report.error(tree.pos, "Class " + tree.name + " not found"); 
        }
        else if (classSym.kind != Symbol.CLASS) {
            tree.type = Type.BAD_TYPE;
            Report.error(tree.pos, "Not a class: " + tree.name); 
        }
        else {
            tree.type = new Type.ObjType(classSym);
            Type.ClassType classType = (Type.ClassType) classSym.type;
            Symbol[] fields = classType.fields.toArray();
            if (tree.args.length != fields.length)
                Report.error(tree.pos, "Wrong number of arguments");
            else
                for (int i = 0; i < fields.length; i++)
                    checkSubType(tree.pos, argTypes[i], fields[i].type);
        }
        tree.sym = classSym;
    }
    
    public void caseNewArray(NewArray tree) {
        Type base = analyze(tree.baseType);
        Type[] args = analyze(tree.args);
        
        if (base == Type.BAD_TYPE) {
            tree.type = Type.BAD_TYPE;
            Report.error(tree.pos, tree.baseType.toString() + " is not a valid type.");
        } else if (!checkSubTypes(args, Type.INT_TYPE)) {
            tree.type = Type.BAD_TYPE;
            Report.error(tree.pos, "array dimensions should be int.");
        } else {
            tree.type = new Type.ArrayType(base, tree.args.length);
        }
    }

    public void caseSelect(Select tree) {
        Type prefixType = analyze(tree.prefix);
        if (!prefixType.isObjType() /*&&
            !prefixType.isArrayType()*/) {
            tree.type = Type.BAD_TYPE;
            Report.error(tree.pos, "Not an object: " + tree.prefix);            
        }
/*        else if (prefixType.isArrayType()) {
            if (!tree.selector.equals("length")) {
                tree.type = Type.BAD_TYPE;
                Report.error("No field " + tree.selector + " in arrays.");
            } else {
                tree.sym = Symbol.LENGTH;
                tree.type = Type.INT_TYPE;
            }
        }*/
        else {
            Type.ObjType objType = (ObjType) (prefixType.isRefType() ? ((Type.RefType)prefixType).baseType : prefixType);
            Type.ClassType classType = (Type.ClassType) objType.clazz.type;
            Symbol field = classType.fields.lookup(tree.selector);
            if (field == null) {
                tree.type = Type.BAD_TYPE;
                Report.error(tree.pos, "No field " + tree.selector + " in class " + objType.clazz.name);
            }
            else {
                tree.sym = field;
                tree.type = field.type;
            }
        }
    }
    
    public void caseIndex(Index tree) {
        Type left = analyze(tree.array);
        Type[] args = analyze(tree.indexes);
        
        if (! isArrayOrRefToArray(left)) {
            tree.type = Type.BAD_TYPE;
            Report.error(tree.pos, "expected array type (found: " + left + ")");
        } else if (!checkSubTypes(args, Type.INT_TYPE)) {
            tree.type = Type.BAD_TYPE;
            Report.error(tree.pos, "indexes should be ints");
        } else {
            Type.ArrayType arrayType = null;
            if (left instanceof Type.RefType) {
                arrayType = (Type.ArrayType)((Type.RefType) left).baseType;
            } else
                arrayType = (Type.ArrayType)left;

            if (arrayType.dims == args.length)
                tree.type = new Type.RefType(arrayType.baseType);
            else {
                tree.type = Type.BAD_TYPE;
                Report.error(tree.pos, "not enough indexes for " + arrayType.dims + "-dimensional array");
            }
        }
    }
    
    private boolean isArrayOrRefToArray(Type t) {
        return (t instanceof Type.ArrayType ||
                ((t instanceof Type.RefType) && 
                        ((Type.RefType)t).baseType instanceof Type.ArrayType));
    }
}
