/*
 * Created on Mar 21, 2005
 *
 * $Id: DotPrinter.java,v 1.3 2005/04/27 13:56:08 dragos Exp $
 */
package einsc;

/*  miscc -- a compiler for misc
 *    Printer to graphviz's "dot" format.
 *
 *    Michel Schinz
 *    $Id: DotPrinter.java,v 1.3 2005/04/27 13:56:08 dragos Exp $
 */

import einsc.Tree.*;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.Iterator;
import java.io.FileNotFoundException;
import java.io.PrintStream;

/**
 * Generate a GraphViz (http://www.graphviz.org/) graph corresponding
 * to the tree.
 */
class DotPrinter implements Visitor {
    private Node self = null;
    private String fileName;
    
    public DotPrinter(String fileName) {
    	this.fileName = fileName;
    }
    
    public void print(Tree tree) throws FileNotFoundException {
        PrintStream out = new PrintStream(fileName);

        nodeFor(tree);
        out.println("digraph misc_program {");

        Iterator<Node> nodeIt = nodes.values().iterator();
        while (nodeIt.hasNext()) {
            Node node = /*(Node)*/nodeIt.next();
            out.println(node + " ["
                        + " label = \"" + node.label + "\""
                        + ", style=filled"
                        + ", fillcolor = \"" + node.color + "\""
                        + " ];");
        }

        Iterator<Edge> edgeIt = edges.iterator();
        while (edgeIt.hasNext()) {
            Edge edge = /*(Edge)*/edgeIt.next();
            assert edge.from != null;
            out.println(edge.from + " -> " + edge.to
                        + "[ label = \"" + edge.label + "\" ];");
        }

        out.println("}");
    }

    public void caseBad(Bad tree) {
        self = addNode(tree, "BAD");
    }

    public void caseProgram(Program tree) {
        self = addNode(tree, "Program");
        child(self, tree.decls);
        child(self, tree.main);
    }

    public void caseFunDef(FunDef tree) {
        self = addNode(tree, "FunDef");
        child(self, tree.restype);
        child(self, tree.params);
        child(self, tree.body);
    }

    public void caseFormal(Formal tree) {
        self = addNode(tree, "Formal(" + tree.name + ")");
        child(self, tree.vartype);
    }

    public void caseFunType(FunType tree) {
        self = addNode(tree, "FunType");
        child(self, tree.paramtypes);
        child(self, tree.restype);
    }
    
    public void caseArrayType(ArrayType tree) {
        self = addNode(tree, "ArrayType[" + tree.dims + "]");
        child(self, tree.baseType);
    }

    public void caseVarDef(VarDef tree) {
        self = addNode(tree, "VarDef");
        child(self, tree.variable);
        child(self, tree.init);
    }

    public void caseWhile(While tree) {
        self = addNode(tree, "While");
        child(self, tree.cond);
        child(self, tree.stats);
    }

    public void caseIf(If tree) {
        self = addNode(tree, "If");
        child(self, tree.cond);
        child(self, tree.thenp);
        child(self, tree.elsep);
    }

    public void caseAssign(Assign tree) {
        self = addNode(tree, "Assign");
        child(self, tree.var);
        child(self, tree.expr);
    }

    public void caseIdent(Ident tree) {
        self = addNode(tree, "Ident(" + tree.name + ")");
    }

    public void caseNullLit(NullLit tree) {
        self = addNode(tree, "NullLit");
    }

    public void caseIntLit(IntLit tree) {
        self = addNode(tree, "IntLit(" + tree.value + ")");
    }

    public void caseApply(Apply tree) {
        self = addNode(tree, "Apply");
        child(self, tree.fun);
        child(self, tree.args);
    }

    public void caseClassDef(ClassDef tree) {
        self = addNode(tree, "ClassDef");
        child(self, tree.fields);
    }

    public void caseIntType(IntType tree) {
    	self = addNode(tree, "IntType");
    }

    public void caseTypeIdent(TypeIdent tree) {
    	self = addNode(tree, "TypeIdent(" + tree.name + ")");
    }

    public void caseDo(Do tree) {
    	self = addNode(tree, "Do");
        child(self, tree.expr);
    }

    public void caseUnop(Unop tree) {
        String opname;
        
        switch (tree.op) {
            case Tree.NOT:
                opname = "NOT";
                break;
            case Tree.NEG:
                opname = "NEG";
                break;
            default:
                throw new java.lang.Error("Unknown unary operation.");
        }
        
    	self = addNode(tree, "Unop(" + opname + ")");
        child(self, tree.expr);
    }

    public void caseBinop(Binop tree) {
    	String opname;
        
        switch (tree.op) {
            case Tree.ADD:
            	opname = "ADD";
                break;
            case Tree.SUB:
                opname = "SUB";
                break;
            case Tree.MUL:
                opname = "MUL";
                break;
            case Tree.DIV:
                opname = "DIV";
                break;
            case Tree.MOD:
                opname = "MOD";
                break;
            case Tree.EQ:
                opname = "EQ";
                break;
            case Tree.NE:
                opname = "NE";
                break;
            case Tree.LT:
                opname = "LT";
                break;
            case Tree.LE:
                opname = "LE";
                break;
            case Tree.GT:
                opname = "GT";
                break;
            case Tree.GE:
                opname = "GE";
                break;
            case Tree.OR:
                opname = "OR";
                break;
            default:
                throw new java.lang.Error("Unknown binop.");            
        }
        
        self = addNode(tree, "Binop(" + opname + ")");
        child(self, tree.left);
        child(self, tree.right);
    }

    public void caseBlock(Block tree) {
        self = addNode(tree, "Block");
        child(self, tree.stats);
        child(self, tree.main, "main");
    }

    public void caseNew(New tree) {
        self = addNode(tree, "New(" + tree.name + ")");
        child(self, tree.args);
    }

    public void caseSelect(Select tree) {
    	self = addNode(tree, "Select(" + tree.selector + ")");
        child(self, tree.prefix);
    }
    
    public void caseIndex(Index tree) {
        self = addNode(tree, "Index");
        child(self, tree.array, "array");
        child(self, tree.indexes);
    }
    
    public void caseNewArray(NewArray tree) {
        self = addNode(tree, "NewArray");
        child(self, tree.baseType, "base");
        child(self, tree.args);        
    }

    //////////////////////////////////////////////////
    
    private Node nodeFor(Tree tree) {
        Node bkpSelf = self;
        tree.apply(this);
        assert self != bkpSelf : "no node generated for " + tree;
        Node nodeForTree = self;
        self.color = WHITE;
        self = bkpSelf;
        return nodeForTree;
    }

    private void child(Node parent, Tree childTree, String name) {
        Node child = nodeFor(childTree);
        assert parent != null;
        assert child != null;
        addEdge(parent, child, name);
    }

    private void child(Node parent, Tree childTree) {
        child(parent, childTree, "");
    }

    private void child(Node parent, Tree[] childrenTrees) {
        for (int i = 0; i < childrenTrees.length; ++i)
            child(parent, childrenTrees[i]);
    }

    //////////////////////////////////////////////////////////////////////

    private HashMap<Tree, Node> nodes = new HashMap<Tree, Node>();
    private LinkedList<Edge> edges = new LinkedList<Edge>();

    private int nodeCounter = 0;

    private Node addNode(Tree tree, String label) {
        assert !nodes.containsKey(tree);
        Node newNode = new Node(++nodeCounter, label);
        nodes.put(tree, newNode);
        return newNode;
    }

    private Edge addEdge(Node from, Node to, String label) {
        Edge newEdge = new Edge(from, to, label);
        edges.add(newEdge);
        return newEdge;
    }

    private static class Color {
        public final double hue, saturation, value;
        public Color(double hue, double saturation, double value) {
            this.hue = hue;
            this.saturation = saturation;
            this.value = value;
        }

        public String toString() {
            return "" + hue + "," + saturation + "," + value;
        }
    }

    private static Color WHITE = new Color(0.0, 0.0, 1.0);

    private static class Node {
        private final int n;
        public final String label;
        public Color color;
        public String toString() { return "n" + n; }
        public Node(int n, String label, Color color) {
            this.n = n;
            this.label = label;
            this.color = color;
        }
        public Node(int n, String label) {
            this(n, label, WHITE);
        }
    }

    private static class Edge {
        public final Node from, to;
        public final String label;
        public Edge(Node from, Node to, String label) {
            this.from = from;
            this.to = to;
            this.label = label;
        }
    }

}
