/*
 * Created on Mar 21, 2005
 *
 * $Id: Main.java,v 1.2 2005/04/27 13:56:08 dragos Exp $
 */
package einsc;

import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.*;

import einsc.backend.*;
import einsc.backend.cfg.CFG;
import einsc.backend.cfg.CFGDotPrinter;
import einsc.backend.cfg.Lineariser;

/**
 * The entry point in the compiler.
 * 
 * @author Iulian Dragos
 */
public class Main {

	public static void main(String[] args) {
        CmdLine cmdLine = new CmdLine();
        
        cmdLine.parse(args);
        try {
            InputStreamReader in = new FileReader(cmdLine.inFile);
            PrintWriter out = (cmdLine.outFile == null) ? 
            	  new PrintWriter(getFileName(cmdLine.inFile) + ".risc")
			    : new PrintWriter(new FileWriter(cmdLine.outFile));

            Report.inFile = cmdLine.inFile;
            Parser parser = new Parser(in);
            Tree tree = parser.parse();
            in.close();
            
//            if (cmdLine.ast.value)
//                new DotPrinter(getFileName(cmdLine.inFile) + ".dot").print(tree);
//
            Analyzer analyzer = new Analyzer();
            analyzer.analyze(tree);
            if (Report.errors > 0)
                System.err.println(Report.errors + " errors found." );
            else {
                if (cmdLine.ast.value)
                    new DotPrinter(getFileName(cmdLine.inFile) + ".dot").print(tree);
                
                Generator generator = new Generator(analyzer);
                CFG[] cfgs = generator.generate((Tree.Program)tree);
                
                if (cmdLine.cfg.value) {
                    CFGDotPrinter printer = new CFGDotPrinter();
                    
                    for (int i = 0; i < cfgs.length; ++i)
                        printer.print(cfgs[i], new PrintWriter(cfgs[i].name + ".dot"));
                }
                Spiller spiller = new Spiller();

                for (int i = 0; i < cfgs.length; ++i) {
					LinearScan linScan = new LinearScan();
					linScan.scan(cfgs[i]);
					Frame frame = new Frame();
					frame.rewriteCalls(cfgs[i]);
					spiller.rewriteSpills(cfgs[i]);

//					String fileName = "cfg" + cfgs[i].name + ".dot";
//					PrintWriter writer = new PrintWriter(new FileWriter(
//							fileName));
//					printer.print(cfgs[i], writer);
//					writer.close();
				}

				Lineariser lin = new Lineariser();
				Code code = lin.lineariseProgram(analyzer.globalScope, cfgs);
				code.write(out);

            }
        } catch (IOException e) {
            Report.fail(e.getMessage());
        }

	}
    
    /** Return the name of this file, without any extension. */
    public static String getFileName(String s) {
    	int splitPoint = s.lastIndexOf('.');
        
        if (splitPoint != -1) {
        	return s.substring(0, splitPoint);
        } else
            return s;
    }
    
    /** Helper class to handle command line arguments */
    public static class CmdLine {
    	private List<Option> options = new LinkedList<Option>();
        public String inFile;
        public String outFile;
        
		/** Print a graph of the Abstract Syntax Tree */
        public Option ast;
        
        /** Print a graph of the control flow graphs. */
        public Option cfg;
        
        public CmdLine() {
            ast = new Option("-ast", "Prints a graph of the Abstract Syntax Tree.");
            options.add(ast);

			cfg = new Option("-cfg", "Prints a set of .dot files for each CFG in the program.");
            options.add(cfg);
		}
        
        /** Parse the command line. */
        public void parse(String[] args) {
        	for (int i = 0; i < args.length; ++i) {
        		if (args[i].startsWith("-")) 
        			handleOption(args[i]);
                else
                    handleFile(args[i]);
            }
            
            if (inFile == null)
                usage();
        }
        
        /** Try to match the given string to the option
         *  that the program knows */
        private void handleOption(String s) {
        	assert s.startsWith("-");
            boolean matched = false;
            
            for (Option o: options) {
            	if (o.name.startsWith(s)) {
                    o.value = true;
                    matched = true;
                    break;
                }
            }
            
            if (!matched) {
            	System.err.println("Don't know what to do with: " + s);
                System.exit(1);
            }
        }
        
        /** Handle an input or output file name */
        private void handleFile(String f) {
        	if (inFile == null)
                inFile = f;
            else if (outFile == null)
                outFile = f;
            else {
            	System.out.println("Too many file names: " + f);
                System.exit(1);
            }
        }
        
        /** Display usage and exit. */
        public void usage() {
        	System.out.println("Usage: java einsc.Main [options] input [output]");
            System.out.println("Options:");
            for (Option o: options) {
            	System.out.println("\t" + o.name + "\t" + o.description);
            }
            System.out.println();
            System.exit(0);
        }
    }
    
    /** A command line option. */
    public static class Option {
    	public boolean value;
        public String name;
        public String description;
        
        public Option(String name, String description) {
        	this.name        = name;
            this.description = description;
            this.value       = false;
        }
    }
}
