/*  einsc -- a compiler for eins
 *    Symbol tables
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Scope.java,v 1.2 2005/04/27 13:56:08 dragos Exp $
 */

package einsc;

import java.util.LinkedList;

/** This class implements symbol table as immutable linked list
 */
public class Scope {

    /** The next outer scope.
     */
    public final Scope outer;

    /** The first symbol in this scope; all other symbols are chained and
     *  can be accessed using the symbol's 'next' link.
     */
    public Symbol first() { return first; }
    private Symbol first;

    /** Creates a new top-level scope.
     */
    public Scope() {
        this(null);
    }

    /** Creates a new nested scope.
     */
    public Scope(Scope outer) {
        this.outer = outer;
    }

    /** Enters a symbol into the scope.
     */
    public void enter(Symbol symbol) {
        Symbol other = lookup(symbol.name);
        if (other != null)
            Report.error(symbol.pos,
                symbol.name + " is already defined at line " +
                Position.line(other.pos));
        symbol.next = first;
        first = symbol;
    }

    /** Tries to find a symbol for the given name in the scope
     *  (including all outer scopes). If this method doesn't find
     *  one, it will return null.
     */
    public Symbol lookup(String name) {
        for (Scope scope = this; scope != null; scope = scope.outer) {
            Symbol symbol = scope.lookupLocal(name);
            if (symbol != null)
                return symbol;
        }
        return null;
    }

    /** Tries to find a symbol for the given name in the local
     *  scope only. If this method doesn't find
     *  one, it will return null.
     */
    private Symbol lookupLocal(String name) {
        for (Symbol symbol = first; symbol != null; symbol = symbol.next)
            if (symbol.name.equals(name))
                return symbol;
        return null;
    }

    /** Returns a textual representation of this scope (and all outer
     *  scopes).
     */
    public String toString() {
        StringBuffer buffer = new StringBuffer();
        buffer.append("{");
        for (Symbol symbol = first; symbol != null; symbol = symbol.next) {
            if (symbol != first)
                buffer.append(", ");
            buffer.append(symbol);
        }
        if (outer != null) {
            if (first != null)
                buffer.append(", ");
            buffer.append(outer);
        }
        buffer.append("}");
        return buffer.toString();
    }

    /** Convert the local symbols of a scope into an array of symbols.
     */
    public Symbol[] toArray() {
        LinkedList<Symbol> syms = new LinkedList<Symbol>();
        for (Symbol sym = first; sym != null; sym = sym.next)
            syms.add(0, sym);
        return syms.toArray(new Symbol[syms.size()]);
    }

}
