/*  einsc -- a compiler for eins
 *    Type representation
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Type.java,v 1.2 2005/04/27 13:56:08 dragos Exp $
 */

package einsc;


/** This class provides a representation for types in eins
 */
public class Type {

    /** This singleton type is used when there is no type associated
     *  with a tree node.
     */
    public static final Type NO_TYPE = new Type();

    /** This type represents an erroneous type (caused by a typing
     *  error).
     */
    public static final Type BAD_TYPE = new Type();

    /** The type of the null object reference.
     */
    public static final Type NULL_TYPE = new Type();

    /** Type Int.
     */
    public static final Type INT_TYPE = new Type();

    /** This method can be used to check if two types are equal. It
     * must be overriden for object types and function types. It will
     * not be used for class types.
     */
    public boolean isSameType(Type that) {
        return
            (this == BAD_TYPE) ||
            (that == BAD_TYPE) ||
            (this == that);
    }

    /** This method checks if the provided type is a super type
     *  of the receiver.
     */
    public boolean isSubType(Type that) {
        return
            (this == BAD_TYPE) ||
            (that == BAD_TYPE) ||
            ((this == NULL_TYPE) && (
                    (that instanceof ObjType) ||
                    (that.isRefType() && ((RefType)that).baseType instanceof ObjType))) ||
            isSameType(that);
    }
    
    public boolean isRefType() {
        return false;
    }
    
    public boolean isArrayType() {
        return false;
    }
    
    public boolean isObjType() {
        return false;
    }
    
    public boolean isFunType() {
        return false;
    }

    /** Return the less specific type of this type and 'that'.
     */
    public Type max(Type that) {
        if (isSubType(that))
            return that;
        if (that.isSubType(this))
            return this;
        return BAD_TYPE;
    }

    /** Returns a textual representation of this type.
     */
    public String toString() {
        if (this == NO_TYPE)
            return "<none>";
        if (this == BAD_TYPE)
            return "<bad>";
        if (this == NULL_TYPE)
            return "Null";
        if (this == INT_TYPE)
            return "Int";
        throw new Error();
    }

    /** Class types.
     */
    public static class ClassType extends Type {
        public final Scope fields;
        
        public ClassType(Scope fields) {
            this.fields = fields;
        }

        public String toString() {
            return "<class type>";
        }
    }

    /** Object types.
     */
    public static class ObjType extends Type {
        public final Symbol clazz;
        
        public ObjType(Symbol clazz) {
            this.clazz = clazz;
        }

        public boolean isSameType(Type that) {
            return (that instanceof ObjType) &&
                (clazz == ((ObjType) that).clazz);
        } 
        
        public boolean isObjType() {
            return true;
        }

        public String toString() {
            return clazz.name;
        }
    }

    /** Representation for function types.
     */
    public static class FunType extends Type {
        public final Type[] argTypes;
        public final Type resType;

        public FunType(Type[] argTypes, Type resType) {
            this.argTypes = argTypes;
            this.resType = resType;
        }

        public boolean isSameType(Type that) {
            return super.isSameType(that) ||
                ((that instanceof FunType) &&
                 sameArgTypes((FunType) that) &&
                 resType.isSameType(((FunType) that).resType));
        }

        private boolean sameArgTypes(FunType funType) {
            if (argTypes.length != funType.argTypes.length)
                return false;
            for (int i = 0; (i < argTypes.length); i++)
                if (!argTypes[i].isSameType(funType.argTypes[i]))
                    return false;
            return true;
        }

        public boolean isFunType() {
            return true;
        }
        
        public String toString() {
            StringBuffer buffer = new StringBuffer();
            buffer.append("(");
            for (int i = 0; i < argTypes.length; ++i) {
                if (i > 0)
                    buffer.append(", ");
                buffer.append(argTypes[i]);
            }
            buffer.append(") => ");
            buffer.append(resType);
            return buffer.toString();
        }
    }

    /** Representation for array types.
     */
    public static class ArrayType extends Type {
        public final Type baseType;
        public final int  dims;
        public ArrayType(Type baseType, int dims) {
            assert dims > 0 : "No 0-dimension arrays are allowed";
            this.baseType = baseType;
            this.dims     = dims;
        }
        
        public String toString() {
            StringBuffer br = new StringBuffer();
            for (int i = 0; i < dims; ++i)
                br.append("[]");
            return baseType instanceof FunType ? "(" + baseType.toString() + ")" + br
                                               : baseType.toString() + br;
        }
        
        /**
         * Two array types are the same if their base types are the same.
         */
        public boolean isSameType(Type that) {
            return (that instanceof ArrayType &&
                    (baseType.isSameType(((ArrayType)that).baseType)) &&
                    dims == ((ArrayType)that).dims);
        }
        
        public boolean isArrayType() {
            return true;
        }
    }
    
    public static class RefType extends Type {
        public final Type baseType;
        
        public RefType(Type baseType) {
            this.baseType = baseType;
        }
        
        public String toString() {
            return "Ref<" + baseType.toString() + ">";
        }
        
        /**
         * Two ref types are the same if their base types are the same.
         */
        public boolean isSameType(Type that) {
            return (that instanceof RefType &&
                    (baseType.isSameType(((RefType)that).baseType)));
        }
        
        /**
         * Reference types are subtypes of their base types.
         */
        public boolean isSubType(Type that) {
            return isSameType(that) || baseType.isSameType(that);
        }
        
        public boolean isRefType() {
            return true;
        }
        
        public boolean isObjType() {
            return baseType.isObjType();
        }
        
        public boolean isFunType() {
            return baseType.isFunType();
        }
        
        public boolean isArrayType() {
            return baseType.isArrayType();
        }
    }
}
