/*  einsc -- a compiler for eins
 *    Support for code generation
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Code.java,v 1.2 2005/04/27 13:56:09 dragos Exp $
 */

package einsc.backend;

import java.io.PrintWriter;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;

import einsc.backend.Instruction;
import einsc.backend.RISC;


public class Code implements RISC {
    //########################################################################
    // Constants

    /** Index of first general-use register. */
    private static final int RC_MIN =  1;
    /** Index of last general-use register. */
    private static final int RC_MAX = 29;
    
    //########################################################################
    // Instance variables

    /** Already emitted instructions */
    private final List<Instruction> code;

    /** Stack frame size */
    private int frameSize;

    //########################################################################
    // Constructor

    public Code() {
        code = new ArrayList<Instruction>();
        frameSize = 0;        
    }
    
    //########################################################################
    // Methods - address handling

    /** Return address of next instruction. */
    public int pc() {
        return WORD_SIZE * code.size();
    }
    

    //########################################################################
    // Methods - frame handling

    /** Increment frame size by given value. */
    public void incFrameSize(int bytes) {
        frameSize += bytes;
    }

    /** Decrement frame size by given value. */
    public void decFrameSize(int bytes) {
        frameSize -= bytes;
        assert frameSize >= 0;
    }

    /** Return current frame size. */
    public int getFrameSize() {
        return frameSize;
    }

    //########################################################################
    // Methods - code emitting

    public void emit(int opcode) {
		emit(opcode, Integer.MIN_VALUE);
	}

	public void emit(int opcode, int c) {
		emit(opcode, Integer.MIN_VALUE, c);
	}

	public void emit(int opcode, int a, int c) {
		emit(opcode, a, Integer.MIN_VALUE, c);
	}

	public void emit(int opcode, int a, int b, int c) {
		code.add(Instruction.make(opcode, a, b, c));
	}

	public void emit(int opcode, String c) {
		code.add(Instruction.make(opcode, Integer.MIN_VALUE, c));
	}

	public void emit(int opcode, int a, String c) {
		code.add(Instruction.make(opcode, a, c));
	}

	public void emit(Instruction i) {
		code.add(i);
	}

	public void addFirst(Instruction i) {
		code.add(0, i);
	}

	public void addLast(Instruction i) {
		code.add(i);
	}

	public void append(Code other) {
		code.addAll(other.code);
	}
    
    public void append(Iterator<Instruction> instructions) {
    	while (instructions.hasNext())
            code.add(instructions.next());
    }

	public List<Instruction> instructions() {
		return Collections.unmodifiableList(code);
	}

    // ########################################################################
    // Methods - program printing.

    /** Print program. */
    public void write(PrintWriter out) {
        int pos = 0;
		for (int i = 0; i < code.size(); i++) {
			Instruction instr = (Instruction) code.get(i);
			if (instr.opcode.code == RISC.LABEL) {
				out.print("          \t");
			} else {
				String label = Integer.toString(pos);
				while (label.length() < 4)
					label = '0' + label;
				out.print("/* " + label + " */\t");
			}
			out.print(instr);
			pos += WORD_SIZE * instr.size();
			out.println();
		}
		out.flush();
    }

}
