/*  einsc -- a compiler for eins
 *    Code generator
 *
 *    Michel Schinz
 *    $Id: Generator.java,v 1.2 2005/04/27 13:56:09 dragos Exp $
 */

package einsc.backend;

import einsc.Analyzer;
import einsc.DefaultVisitor;
import einsc.Position;
import einsc.Symbol;
import einsc.Tree;
import einsc.Type;
import einsc.Visitor;
import einsc.Tree.*;
import einsc.backend.cfg.CFG;
import java.util.*;

public class Generator {
    private final Analyzer analyzer;

    private final Gen genVisitor;
    private final GenLoad genLoadVisitor;
    private final GenCond genCondVisitor;
    private final RegisterPool regPool;
    private List<CFG> cfgs;
    
    private CFG currentCFG;
    private CFG.Node currentNode;
    
    /** frame management */
    private int frameSize;
    
    private void incFrameSize(int n) {
    	frameSize += n;
    }
    
    private void decFrameSize(int n) {
    	assert frameSize >= n;
        frameSize -= n;
    }
    
    private int getFrameSize() {
    	return frameSize;
    }

    public Generator(Analyzer analyzer) {
        this.analyzer = analyzer;
        this.genVisitor = new Gen();
        this.genLoadVisitor = new GenLoad();
        this.genCondVisitor = new GenCond();
        this.regPool = new RegisterPool(32, Integer.MAX_VALUE);
        this.cfgs = new LinkedList<CFG>();
    }
    
    public CFG[] generate(Tree.Program tree) {
        gen(tree);
        return cfgs.toArray(new CFG[cfgs.size()]);
    }
    
    /** Return a new CFG and add it in the list. */
    protected CFG addNewCFG(String name) {
    	CFG c = new CFG(name);
        cfgs.add(c);
        return c;
    }
    
    public void gen(Tree tree) {
        genVisitor.generate(tree);
    }

    public void gen(Tree[] trees) {
        for (int i = 0; i < trees.length; ++i)
            gen(trees[i]);
    }

    public void genLoad(Tree tree, int targetReg) {
        assert regPool.used().get(targetReg)
            : "invariant violated: target register R" + targetReg
            + " not allocated";
        genLoadVisitor.generate(tree, targetReg, false); // default is to load values
    }
	
	public void genLoad(Tree tree, int targetReg, boolean loadAddress) {
        assert regPool.used().get(targetReg)
        	: "invariant violated: target register R" + targetReg
        	+ " not allocated";
		genLoadVisitor.generate(tree, targetReg, loadAddress);
	}

    public void genCond(Tree tree, CFG.Node contTrue, CFG.Node contFalse) {
        genCondVisitor.generate(tree, contTrue, contFalse);
    }

    /**
     * Visitor generating code for statements, in a general sense,
     * that is all constructs which do not have a value.
     */
    private class Gen extends DefaultVisitor implements Visitor, RISC {

        public void generate(Tree tree) {
            tree.apply(this);
        }

        public void caseProgram(Program tree) {

            // generate predefined classes
            Symbol listClass = analyzer.globalScope.lookup("List");
            Symbol[] fields = ((Type.ClassType) listClass.type).fields.toArray();
            for (int i = 0; i < fields.length; i++)
                fields[i].offset = i * WORD_SIZE;

            // generate class and function definitions
            gen(tree.decls);

            currentCFG = addNewCFG("_Main");
            currentNode = currentCFG.addNewNode();
            genLoad(tree.main, regPool.get());

            // exit
            currentNode.emit(RET, ZERO);
        }

        public void caseClassDef(ClassDef tree) {
            for(int i = 0; i < tree.fields.length; i++)
                ((Formal) tree.fields[i]).sym.offset = i * WORD_SIZE;
        }

        public void caseFunDef(FunDef tree) {
            tree.sym.label = "__" + tree.sym.name;
            currentCFG = addNewCFG(tree.sym.label);
            currentNode = currentCFG.addNewNode();
            
            gen(tree.params);

            incFrameSize(WORD_SIZE);
            currentNode.emit(PSH, LNK, SP, WORD_SIZE);
            
            int resReg = regPool.get();
            genLoad(tree.body, resReg);
            currentNode.emit(ADDI, RES, resReg, 0);
            regPool.free(resReg);
            
            int popSize = WORD_SIZE * (1 + tree.params.length);
            currentNode.emit(POP, LNK, SP, popSize);
            decFrameSize(popSize);
            assert getFrameSize() == 0
                : "non-empty frame (size " + getFrameSize() + ")";
            currentNode.emit(RET, LNK);
        }

        public void caseFormal(Formal tree) {
            incFrameSize(WORD_SIZE);
            tree.sym.offset = getFrameSize();
        }

        public void caseWhile(While tree) {
            CFG.Node condNode = currentCFG.addNewNode();
            currentCFG.addNewEdge(currentNode, condNode, BEQ, ZERO);
            
            CFG.Node statsNode = currentCFG.addNewNode();
            CFG.Node afterNode = currentCFG.addNewNode();
            
            currentNode = condNode;
            genCond(tree.cond, statsNode, afterNode);
            
            int beginFrameSize = getFrameSize();
            currentNode = statsNode;
            gen(tree.stats);
            int deltaFrameSize = getFrameSize() - beginFrameSize;
            if (deltaFrameSize > 0) {
                currentNode.emit(ADDI, SP, SP, deltaFrameSize);
                decFrameSize(deltaFrameSize);
            }
			currentCFG.addNewEdge(currentNode, condNode, BEQ, ZERO);

            currentNode = afterNode;
        }

        public void caseVarDef(VarDef tree) {
            int tmpReg = regPool.get();
            genLoad(tree.init, tmpReg);
            gen(tree.variable);
            currentNode.emit(PSH, tmpReg, SP, WORD_SIZE);
        }

        public void caseAssign(Assign tree) {
            int exprReg = regPool.get();
			int locReg  = regPool.get();
			
			genLoad(tree.var, locReg, true); // load address
            genLoad(tree.expr, exprReg); // load value
			
			// !!
            //int realOffset = getFrameSize() - tree.sym.offset;
			
            currentNode.emit(STW, exprReg, locReg, 0);
        }

        public void caseDo(Do tree) {
            int tmpReg = regPool.get();
            genLoad(tree.expr, tmpReg);
        }

		public void caseDefault(Tree tree) {
            throw new Error("unexpected case");
        }
    }

    /**
     * Visitor generating code for expressions, that is constructs
     * which produce a value.
     */
    private class GenLoad extends DefaultVisitor implements Visitor, RISC {
        private int targetReg;
		private boolean loadAddress;

        public void generate(Tree tree, int targetReg, boolean loadAddress) {
            int bkpTargetReg = this.targetReg;
			boolean bkpLoadAddress = this.loadAddress;
			
            this.targetReg = targetReg;
			this.loadAddress = loadAddress;
			
            tree.apply(this);
            
			this.targetReg = bkpTargetReg;
			this.loadAddress = bkpLoadAddress;
        }
		
		public void generate(Tree tree, int targetReg) {
			generate(tree, targetReg, loadAddress);
		}

        public void caseIf(If tree) {
            CFG.Node thenNode = currentCFG.addNewNode();
            CFG.Node elseNode = currentCFG.addNewNode();
            CFG.Node contNode = currentCFG.addNewNode();
            
            genCond(tree.cond, thenNode, elseNode);

            currentNode = thenNode;
            genLoad(tree.thenp, targetReg);
            currentCFG.addNewEdge(currentNode, contNode, BEQ, ZERO);
			
            currentNode = elseNode;
            genLoad(tree.elsep, targetReg);
            currentCFG.addNewEdge(currentNode, contNode, BEQ, ZERO);
            
            currentNode = contNode;
        }

        public void caseBinop(Binop tree) {
            switch (tree.op) {
            case Tree.ADD: case Tree.SUB:
            case Tree.MUL: case Tree.DIV: case Tree.MOD: {
                int leftReg = regPool.get();
                genLoad(tree.left, leftReg);
                int rightReg = regPool.get();
                genLoad(tree.right, rightReg);
            	currentNode.emit(opcode(tree.op), targetReg, leftReg, rightReg);
            } break;
            default:
                caseDefault(tree);
            }
        }

        public void caseIdent(Ident tree) {			
            if (tree.sym.kind == Symbol.VAR) {
				if (loadAddress) {
					int realOffset = getFrameSize() - tree.sym.offset;
					currentNode.emit(ADDIU, targetReg, SP, realOffset);
				} else {
					int realOffset = getFrameSize() - tree.sym.offset;
					currentNode.emit(LDW, targetReg, SP, realOffset);
				}
            } else {
                assert tree.sym.kind == Symbol.FUNC;
                currentNode.emit(LDA, targetReg, tree.sym.label);
            }
        }

        public void caseNullLit(NullLit tree) {
            emitLoadConstant(targetReg, 0);
        }

        public void caseIntLit(IntLit tree) {
            emitLoadConstant(targetReg, tree.value);
        }

        public void caseApply(Apply tree) {
            String funLabel = null;
            int funReg = -1;
            
            if ((tree.fun instanceof Ident)
                && ((Ident)tree.fun).sym.kind == Symbol.FUNC)
                funLabel = ((Ident)tree.fun).sym.label;
            else {
                funReg = regPool.get();
                genLoad(tree.fun, funReg);
            }

            // load arguments
            List<Integer> args = new LinkedList<Integer>();
            for (int i = 0; i < tree.args.length; ++i) {
                int tmpReg = regPool.get();
                genLoad(tree.args[i], tmpReg);
                args.add(new Integer(tmpReg));
            }

            if (funLabel != null)
                currentNode.emit(new PJSR(targetReg, args, funLabel));
            else {
                currentNode.emit(new PCALL(targetReg, args, funReg));
            }
        }

        public void caseBlock(Block tree) {
            int beginFrameSize = getFrameSize();
            gen(tree.stats);
            genLoad(tree.main, targetReg);
            int deltaFrameSize = getFrameSize() - beginFrameSize;
            if (deltaFrameSize > 0) {
                currentNode.emit(ADDI, SP, SP, deltaFrameSize);
                decFrameSize(deltaFrameSize);
            }
        }

        public void caseNew(New tree) {
            int size = tree.args.length;
            emitLoadConstant(targetReg, size * WORD_SIZE);
            currentNode.emit(SYSCALL, targetReg, targetReg, SYS_GC_ALLOC);
            int tmpReg = regPool.get();
            for (int i = 0; i < size; ++i) {
                genLoad(tree.args[i], tmpReg);
                currentNode.emit(STW, tmpReg, targetReg, i * WORD_SIZE);
            }
        }

        public void caseSelect(Select tree) {
            genLoad(tree.prefix, targetReg);
            
            CFG.Node errorNode = makeErrorBlock(tree.pos);
            CFG.Node contNode = currentCFG.addNewNode();
            
            currentCFG.addNewEdge(currentNode, errorNode, BEQ, targetReg);
            currentCFG.addNewEdge(currentNode, contNode, BNE, targetReg);
            
            currentNode = contNode;
            currentNode.emit(LDW, targetReg, targetReg, tree.sym.offset);
        }

		private CFG.Node makeErrorBlock(int pos) {
			CFG.Node errorNode = currentCFG.addNewNode();
			
			// slight change: instead of exit code 1, return the line where
			// the null pointer dereference appeared.
            errorNode.emit(ORIU, targetReg, ZERO, Position.line(pos));
			errorNode.emit(SYSCALL, targetReg, ZERO, SYS_EXIT);
			
			return errorNode;
		}
		
		@Override
		public void caseIndex(Index tree) {
			int arrReg = regPool.get();
			genLoad(tree.array, arrReg);

            CFG.Node errorNode = makeErrorBlock(tree.pos);
            CFG.Node contNode;
			
			// compute indexes
			int[] idx = new int[tree.indexes.length];
			for (int i = 0; i < tree.indexes.length; ++i) {
				idx[i] = regPool.get();
				genLoad(tree.indexes[i], idx[i]);
			}
			
			int dims[] = new int[tree.indexes.length];
			
			// check array bounds
			for (int i = 0; i < idx.length; ++i) {
				dims[i] = regPool.get();
				currentNode.emit(LDW, dims[i], arrReg, i * WORD_SIZE);
				currentNode.emit(CMP, targetReg, dims[i], idx[i]);
				currentCFG.addNewEdge(currentNode, errorNode, BLE, targetReg);
				contNode = currentCFG.addNewNode();
				
				currentCFG.addNewEdge(currentNode, contNode, BGT, targetReg);
				currentNode = contNode;
				currentCFG.addNewEdge(currentNode, errorNode, BLT, idx[i]);
				contNode = currentCFG.addNewNode();
				currentCFG.addNewEdge(currentNode, contNode, BGE, idx[i]);
				currentNode = contNode;
			}

			int offsetReg = regPool.get();
			
			// compute memory location
			currentNode.emit(ADDIU, arrReg, arrReg, idx.length * WORD_SIZE); 
			for (int i = 0; i < idx.length; ++i) {
				currentNode.emit(ORIU, offsetReg, idx[i], 0);
				for (int j = i + 1; j < dims.length; ++j) 
					currentNode.emit(MUL, offsetReg, offsetReg, dims[j]);
				// word-align the address
				currentNode.emit(MULIU, offsetReg, offsetReg, WORD_SIZE);
				currentNode.emit(ADD, arrReg, arrReg, offsetReg);
			}
			
			if (loadAddress)
				currentNode.emit(ORIU, targetReg, arrReg, 0);
			else
				currentNode.emit(LDW, targetReg, arrReg, 0);
		}

        public void caseUnop(Unop tree) {
            if (tree.op == Tree.NEG) {
                genLoad(tree.expr, targetReg);
                currentNode.emit(SUB, targetReg, ZERO, targetReg);
            }
            else
                caseDefault(tree);
        }
		
		@Override
		public void caseNewArray(NewArray tree) {
			int[] sizes = new int[tree.args.length];
			
			// load array sizes
			for (int i = 0; i < tree.args.length; ++i) {
				sizes[i] = regPool.get();
				genLoad(tree.args[i], sizes[i]);
			}
			
			currentNode.emit(ORIU, targetReg, sizes[0], 0);
			for (int i = 1; i < tree.args.length; ++i)
				currentNode.emit(MUL, targetReg, targetReg, sizes[i]);
			currentNode.emit(ADDI, targetReg, targetReg, tree.args.length);
			currentNode.emit(MULI, targetReg, targetReg, WORD_SIZE);
			
			// allocate memory (guaranteed to be 0-initialized)
			currentNode.emit(SYSCALL, targetReg, targetReg, SYS_GC_ALLOC);
			
			// initialize the array
			for (int i = 0; i < tree.args.length; ++i) 
				currentNode.emit(STW, sizes[i], targetReg, i * WORD_SIZE);
		}

        public void caseDefault(Tree tree) {
            genLoad(new If(tree.pos,
                           tree,
                           new IntLit(tree.pos, 1),
                           new IntLit(tree.pos, 0)),
                    targetReg);
        }

        private void emitLoadConstant(int reg, int value) {
            int highBits = value >> 16;
            int lowBits = value & 0xFFFF;
            if (highBits != 0) {
                currentNode.emit(ORIU, reg, ZERO, highBits);
                currentNode.emit(LSHI, reg, reg, 16);
                currentNode.emit(ORIU, reg, reg, lowBits);
            } else
                currentNode.emit(ORIU, reg, ZERO, lowBits);
        }
        private int opcode(int op) {
            switch (op) {
            case Tree.ADD: return ADD;
            case Tree.SUB: return SUB;
            case Tree.MUL: return MUL;
            case Tree.DIV: return DIV;
            case Tree.MOD: return MOD;
            default: throw new Error("unknown opcode: " + op);
            }
        }
    }

    /**
     * Visitor generating code for conditions, that is constructs
     * which jump somewhere on the result of a test.
     */
    private class GenCond extends DefaultVisitor implements Visitor, RISC {
        private CFG.Node trueCont;
        private CFG.Node falseCont;

        public void generate(Tree tree, CFG.Node trueCont, CFG.Node falseCont) {
            CFG.Node bkpTrueCont = this.trueCont;
            this.trueCont = trueCont;
            
            CFG.Node bkpFalseCont = this.falseCont;
            this.falseCont = falseCont;
            
            tree.apply(this);
            this.trueCont = bkpTrueCont;
            this.falseCont = bkpFalseCont;
        }

        public void caseBinop(Binop tree) {
            switch (tree.op) {
            case Tree.EQ: case Tree.NE: case Tree.LT:
            case Tree.LE: case Tree.GT: case Tree.GE:
                int leftReg = regPool.get();
                genLoad(tree.left, leftReg);
                int rightReg = regPool.get();
                genLoad(tree.right, rightReg);
				int targetReg = regPool.get();
                
            	currentNode.emit(CMP, targetReg, leftReg, rightReg);
                
                currentCFG.addNewEdge(currentNode, 
                        trueCont, 
                        opcode(tree.op, true), 
                        targetReg);                
                currentCFG.addNewEdge(currentNode,
                        falseCont,
                        opcode(tree.op, false),
                        targetReg);
                break;
            case Tree.OR:
                CFG.Node rightNode = currentCFG.addNewNode();
                genCond(tree.left, trueCont, rightNode);
                currentNode = rightNode;
                genCond(tree.right, trueCont, falseCont);
                break;
            default:
                caseDefault(tree);
            }
        }

        public void caseUnop(Unop tree) {
            if (tree.op == Tree.NOT)
                genCond(tree.expr, falseCont, trueCont);
            else
                caseDefault(tree);
        }

        public void caseDefault(Tree tree) {
            int tmpReg = regPool.get();
            genLoad(tree, tmpReg);
            currentCFG.addNewEdge(currentNode, trueCont, BNE, tmpReg);
            currentCFG.addNewEdge(currentNode, falseCont, BEQ, tmpReg);            
        }

        private int opcode(int op, boolean negated) {
            switch (op) {
            case Tree.EQ : return (negated ? BEQ : BNE);
            case Tree.NE : return (negated ? BNE : BEQ);
            case Tree.LT : return (negated ? BLT : BGE);
            case Tree.LE : return (negated ? BLE : BGT);
            case Tree.GT : return (negated ? BGT : BLE);
            case Tree.GE : return (negated ? BGE : BLT);
            default      : throw new Error("invalid operator: " + op);
            }
        }
    }
}
