package einsc.backend;

import java.util.BitSet;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;

public abstract class Instruction implements RISC {
    public final Opcode opcode;

    protected int a, b, c;
    public boolean sideEffect() {
        // If the stack pointer is modified there is a side effect.
        return defs().get(RISC.SP) || opcode.sideEffect;
    }

    protected Instruction(Opcode opcode, int a, int b, int c) {
        this.opcode = opcode;
        this.a = a;
        this.b = b;
        this.c = c;
    }

    public static Instruction make(int opcode, int a, int b, int c) {
        switch (opcode) {
        case ADD:     return new   ADD(a,b,c);
        case SUB:     return new   SUB(a,b,c);
        case MUL:     return new   MUL(a,b,c);
        case DIV:     return new   DIV(a,b,c);
        case MOD:     return new   MOD(a,b,c);
        case CMP:     return new   CMP(a,b,c);
        case OR:      return new    OR(a,b,c);
        case AND:     return new   AND(a,b,c);
        case BIC:     return new   BIC(a,b,c);
        case XOR:     return new   XOR(a,b,c);
        case LSH:     return new   LSH(a,b,c);
        case ASH:     return new   ASH(a,b,c);
        case CHK:     return new   CHK(a,  c);
        case ADDI:    return new  ADDI(a,b,c);
        case SUBI:    return new  SUBI(a,b,c);
        case MULI:    return new  MULI(a,b,c);
        case DIVI:    return new  DIVI(a,b,c);
        case MODI:    return new  MODI(a,b,c);
        case CMPI:    return new  CMPI(a,b,c);
        case ORI:     return new   ORI(a,b,c);
        case ANDI:    return new  ANDI(a,b,c);
        case BICI:    return new  BICI(a,b,c);
        case XORI:    return new  XORI(a,b,c);
        case LSHI:    return new  LSHI(a,b,c);
        case ASHI:    return new  ASHI(a,b,c);
        case CHKI:    return new  CHKI(a,  c);
        case ADDIU:   return new ADDIU(a,b,c);
        case SUBIU:   return new SUBIU(a,b,c);
        case MULIU:   return new MULIU(a,b,c);
        case DIVIU:   return new DIVIU(a,b,c);
        case MODIU:   return new MODIU(a,b,c);
        case CMPIU:   return new CMPIU(a,b,c);
        case ORIU:    return new  ORIU(a,b,c);
        case ANDIU:   return new ANDIU(a,b,c);
        case BICIU:   return new BICIU(a,b,c);
        case XORIU:   return new XORIU(a,b,c);
        case CHKIU:   return new CHKIU(a,  c);
        case LDW:     return new   LDW(a,b,c);
        case LDB:     return new   LDB(a,b,c);
        case POP:     return new   POP(a,b,c);
        case STW:     return new   STW(a,b,c);
        case STB:     return new   STB(a,b,c);
        case PSH:     return new   PSH(a,b,c);
        case RET:     return new   RET(    c);
        case CALL:    return new  CALL(    c);
        case BREAK:   return new BREAK(     );
        case SYSCALL: return new SYSCALL(a,b,c);
        default:      throw new IllegalArgumentException();
        }
    }

    public static Instruction make(int opcode, int a, String c) {
        switch (opcode) {
        case LDA:   return new   LDA(a,  c);
        case BEQ:   return new   BEQ(a,  c);
        case BNE:   return new   BNE(a,  c);
        case BLT:   return new   BLT(a,  c);
        case BGE:   return new   BGE(a,  c);
        case BLE:   return new   BLE(a,  c);
        case BGT:   return new   BGT(a,  c);
        case BSR:   return new   BSR(    c);
        case JSR:   return new   JSR(    c);
        case LABEL: return new LABEL(    c);
        default:    throw new IllegalArgumentException();
        }
    }

    /** The size, in words, of this instruction. */
    abstract public int size();

    /** Set of registers defined by this instruction. */
    public BitSet defs() {
        assert opcode.defs != Opcode.REG_UNKNOWN;

        BitSet defs = new BitSet();
        if ((opcode.defs & Opcode.REG_A) != 0) defs.set(a);
        if ((opcode.defs & Opcode.REG_B) != 0) defs.set(b);
        if ((opcode.defs & Opcode.REG_C) != 0) defs.set(c);
        if ((opcode.defs & Opcode.REG_LNK) != 0) defs.set(RISC.LNK);
        return defs;
    }

    /** Set of registers used by this instruction. */
    public BitSet uses() {
        assert opcode.uses != Opcode.REG_UNKNOWN;

        BitSet uses = new BitSet();
        if ((opcode.uses & Opcode.REG_A) != 0) uses.set(a);
        if ((opcode.uses & Opcode.REG_B) != 0) uses.set(b);
        if ((opcode.uses & Opcode.REG_C) != 0) uses.set(c);
        if ((opcode.uses & Opcode.REG_LNK) != 0) uses.set(RISC.LNK);
        return uses;
    }

    public void replaceRegs(int[] map) {
        assert opcode.defs != Opcode.REG_UNKNOWN : this;
        assert opcode.uses != Opcode.REG_UNKNOWN : this;

        int allRegs = opcode.uses | opcode.defs;
        if ((allRegs & Opcode.REG_A) != 0 && a >= 0 && a < map.length)
            a = map[a];
        if ((allRegs & Opcode.REG_B) != 0 && b >= 0 && b < map.length)
            b = map[b];
        if ((allRegs & Opcode.REG_C) != 0 && c >= 0 && c < map.length)
            c = map[c];
    }


    public String toString() {
        return Helper.pad(7, opcode.mnemonic, ' ') + "\t" + argsToString();
    }

    protected String regToString(int reg) {
        return (reg > 31 ? "t" : "r")
            + Helper.pad(2, '0', String.valueOf(reg));
    }

    abstract protected String argsToString();
}

/** Super-class for real instructions, which all have a size of 1. */
abstract class RealInstruction extends Instruction {
    protected RealInstruction(Opcode opcode, int a, int b, int c) {
        super(opcode, a, b, c);
    }

    public int size() { return 1; }
}

//########################################################################
// Opcodes - instructions with registers

/**
 * <b><code>BinOp a b c</code></b><br>
 *    <code>R.a = R.b BinOp R.c</code>
 */

abstract class OpRaRbRc extends RealInstruction {
    public OpRaRbRc(Opcode opcode, int a, int b, int c) {
        super(opcode, a, b, c);
    }
    protected String argsToString() {
        return regToString(a) + ", " + regToString(b) + ", " + regToString(c);
    }
}

/**
 * <p>Opcode: integer addition.</p>
 * <b><code>ADD a b c</code></b><br>
 *    <code>R.a = R.b + R.c</code>
 */
class ADD extends OpRaRbRc {
    public ADD(int a, int b, int c) { super(Opcode.ADD, a, b, c); }
}

/**
 * <p>Opcode: integer subtraction.</p>
 * <b><code>SUB a b c</code></b><br>
 *    <code>R.a = R.b - R.c</code>
 */
class SUB extends OpRaRbRc {
    public SUB(int a, int b, int c) { super(Opcode.SUB, a, b, c); }
}

/**
 * <p>Opcode: integer multiplication.</p>
 * <b><code>MUL a b c</code></b><br>
 *    <code>R.a = R.b * R.c</code>
 */
class MUL extends OpRaRbRc {
    public MUL(int a, int b, int c) { super(Opcode.MUL, a, b, c); }
}

/**
 * <p>Opcode: integer division.</p>
 * <b><code>DIV a b c</code></b><br>
 *    <code>R.a = R.b / R.c</code>
 */
class DIV extends OpRaRbRc {
    public DIV(int a, int b, int c) { super(Opcode.DIV, a, b, c); }
}

/**
 * <p>Opcode: integer modulo.</p>
 * <b><code>MOD a b c</code></b><br>
 *    <code>R.a = R.b % R.c</code>
 */
class MOD extends OpRaRbRc {
    public MOD(int a, int b, int c) { super(Opcode.MOD, a, b, c); }
}

/**
 * <p>Opcode: integer comparison.</p>
 * <b><code>CMP a b c</code></b><br>
 *    <code>R.a = &lt;a value with the same sign as (R.b - R.c)
 *    but with a possibly different magnitude&gt;</code>
 */
class CMP extends OpRaRbRc {
    public CMP(int a, int b, int c) { super(Opcode.CMP, a, b, c); }
}

/**
 * <p>Opcode: logical or.</p>
 * <b><code>OR a b c</code></b><br>
 *    <code>R.a = R.b | R.c</code>
 */
class OR extends OpRaRbRc {
    public OR(int a, int b, int c) { super(Opcode.OR, a, b, c); }
}

/**
 * <p>Opcode: logical and.</p>
 * <b><code>AND a b c</code></b><br>
 *    <code>R.a = R.b & R.c</code>
 */
class AND extends OpRaRbRc {
    public AND(int a, int b, int c) { super(Opcode.AND, a, b, c); }
}

/**
 * <p>Opcode: logical bic.</p>
 * <b><code>BIC a b c</code></b><br>
 *    <code>R.a = R.b & ~R.c</code>
 */
class BIC extends OpRaRbRc {
    public BIC(int a, int b, int c) { super(Opcode.BIC, a, b, c); }
}

/**
 * <p>Opcode: logical xor.</p>
 * <b><code>XOR a b c</code></b><br>
 *    <code>R.a = R.b ^ R.c</code>
 */
class XOR extends OpRaRbRc {
    public XOR(int a, int b, int c) { super(Opcode.XOR, a, b, c); }
}

/**
 * <p>Opcode: logical shift.</p>
 * <b><code>LSH a b c</code></b><br>
 *    <code>R.a = (R.c &gt; 0) ?
 *      (R.b &lt;&lt; R.c) : (R.b &gt;&gt;&gt; -R.c)</code>
 */
class LSH extends OpRaRbRc {
    public LSH(int a, int b, int c) { super(Opcode.LSH, a, b, c); }
}

/**
 * <p>Opcode: arithmetic shift.</p>
 * <b><code>ASH a b c</code></b><br>
 *    <code>R.a = (R.c &gt; 0) ?
 *      (R.b &lt;&lt; R.c) : (R.b &gt;&gt; -R.c)</code>
 */
class ASH extends OpRaRbRc {
    public ASH(int a, int b, int c) { super(Opcode.ASH, a, b, c); }
}


abstract class OpRa__Rc extends RealInstruction {
    public OpRa__Rc(Opcode opcode, int a, int c) {
        super(opcode, a, Integer.MIN_VALUE, c);
        this.a = a;
        this.c = c;
    }
    protected String argsToString() {
        return regToString(a) + ", " + regToString(c);
    }
}

/**
 * <p>Opcode: bound check.</p>
 * <b><code>CHK a c</code></b><br>
 *    raise an error if not <code>(0 &lt;= R.a &lt; R.c)</code>
 */
class CHK extends OpRa__Rc {
    public CHK(int a, int c) { super(Opcode.CHK, a, c); }
}



//########################################################################

/**
 * <b><code>BinOpI a b ic</code></b><br>
 *    <code>R.a = R.b BinOpI ic</code>
 */
abstract class OpRaRbxc extends RealInstruction {
    public OpRaRbxc(Opcode opcode, int a, int b, int c) {
        super(opcode, a, b, c);
    }
}

abstract class OpRaRbIc extends OpRaRbxc {
    public OpRaRbIc(Opcode opcode, int a, int b, int c) {
        super(opcode, a, b, c);
    }
    protected String argsToString() {
        return regToString(a) + ", " + regToString(b) + ", " + c;
    }
}

//########################################################################
// Opcodes - instructions with register and signed immediate

/**
 * <p>Opcode: integer addition with signed immediate.</p>
 * <b><code>ADDI a b ic</code></b><br>
 *    <code>R.a = R.b + ic</code>
 */
class ADDI extends OpRaRbIc {
    public ADDI(int a, int b, int c) { super(Opcode.ADDI, a, b, c); }
}

/**
 * <p>Opcode: integer subtraction with signed immediate.</p>
 * <b><code>SUBI a b ic</code></b><br>
 *    <code>R.a = R.b - ic</code>
 */
class SUBI extends OpRaRbIc {
    public SUBI(int a, int b, int c) { super(Opcode.SUBI, a, b, c); }
}

/**
 * <p>Opcode: integer multiplication with signed immediate.</p>
 * <b><code>MULI a b ic</code></b><br>
 *    <code>R.a = R.b * ic</code>
 */
class MULI extends OpRaRbIc {
    public MULI(int a, int b, int c) { super(Opcode.MULI, a, b, c); }
}

/**
 * <p>Opcode: integer division with signed immediate.</p>
 * <b><code>DIVI a b ic</code></b><br>
 *    <code>R.a = R.b / ic</code>
 */
class DIVI extends OpRaRbIc {
    public DIVI(int a, int b, int c) { super(Opcode.DIVI, a, b, c); }
}

/**
 * <p>Opcode: integer modulo with signed immediate.</p>
 * <b><code>MODI a b ic</code></b><br>
 *    <code>R.a = R.b % ic</code>
 */
class MODI extends OpRaRbIc {
    public MODI(int a, int b, int c) { super(Opcode.MODI, a, b, c); }
}

/**
 * <p>Opcode: integer comparison with signed immediate.</p>
 * <b><code>CMPI a b ic</code></b><br>
 *    <code>R.a = &lt;a value with the same sign as (R.b - ic)
 *    but with a possibly different magnitude&gt;</code>
 */
class CMPI extends OpRaRbIc {
    public CMPI(int a, int b, int c) { super(Opcode.CMPI, a, b, c); }
}

/**
 * <p>Opcode: logical or with signed immediate.</p>
 * <b><code>ORI a b ic</code></b><br>
 *    <code>R.a = R.b | ic</code>
 */
class ORI extends OpRaRbIc {
    public ORI(int a, int b, int c) { super(Opcode.ORI, a, b, c); }
}

/**
 * <p>Opcode: logical and with signed immediate.</p>
 * <b><code>ANDI a b ic</code></b><br>
 *    <code>R.a = R.b & ic</code>
 */
class ANDI extends OpRaRbIc {
    public ANDI(int a, int b, int c) { super(Opcode.ANDI, a, b, c); }
}

/**
 * <p>Opcode: logical bic with signed immediate.</p>
 * <b><code>BICI a b ic</code></b><br>
 *    <code>R.a = R.b & ~ic</code>
 */
class BICI extends OpRaRbIc {
    public BICI(int a, int b, int c) { super(Opcode.BICI, a, b, c); }
}


/**
 * <p>Opcode: logical xor with signed immediate.</p>
 * <b><code>XORI a b ic</code></b><br>
 *    <code>R.a = R.b ^ ic</code>
 */
class XORI extends OpRaRbIc {
    public XORI(int a, int b, int c) { super(Opcode.XORI, a, b, c); }
}

/**
 * <p>Opcode: logical shift with signed immediate.</p>
 * <b><code>LSHI a b ic</code></b><br>
 *    <code>R.a = (ic &gt; 0) ?
 *      (R.b &lt;&lt; ic) : (R.b &gt;&gt;&gt; -ic)</code>
 */
class LSHI extends OpRaRbIc {
    public LSHI(int a, int b, int c) { super(Opcode.LSHI, a, b, c); }
}

/**
 * <p>Opcode: arithmetic shift with signed immediate.</p>
 * <b><code>ASHI a b ic</code></b><br>
 *    <code>R.a = (ic &gt; 0) ?
 *      (R.b &lt;&lt; ic) : (R.b &gt;&gt; -ic)</code>
 */
class ASHI extends OpRaRbIc {
    public ASHI(int a, int b, int c) { super(Opcode.ASHI, a, b, c); }
}

abstract class OpRa__Ic extends RealInstruction {
    public OpRa__Ic(Opcode opcode, int a, int c) {
        super(opcode, a, Integer.MIN_VALUE, c);
    }
    protected String argsToString() {
        return regToString(a) + ", " + c;
    }
}

/**
 * <p>Opcode: bound check with signed immediate.</p>
 * <b><code>CHKI a ic</code></b><br>
 *    raise an error if not <code>(0 &lt;= R.a < ic)</code>
 */
class CHKI extends OpRa__Ic {
    public CHKI(int a, int c) { super(Opcode.CHKI, a, c); }
}


//########################################################################
/**
 * <b><code>BinOpIU a b uc</code></b><br>
 *    <code>R.a = R.b BinOpUI uc</code>
 */

abstract class OpRaRbUc extends OpRaRbxc {
    public OpRaRbUc(Opcode opcode, int a, int b, int c) {
        super(opcode, a, b, c);
    }
    protected String argsToString() {
        return regToString(a) + ", " + regToString(b) + ", " + (c & 0xFFFF);
    }
}

//########################################################################
// Opcodes - instructions with register and unsigned immediate

/**
 * <p>Opcode: integer addition with unsigned immediate.</p>
 * <b><code>ADDIU a b uc</code></b><br>
 *    <code>R.a = R.b + uc</code>
 */
class ADDIU extends OpRaRbUc {
    public ADDIU(int a, int b, int c) { super(Opcode.ADDIU, a, b, c); }
}

class SUBIU extends OpRaRbUc {
    public SUBIU(int a, int b, int c) { super(Opcode.SUBIU, a, b, c); }
}

class MULIU extends OpRaRbUc {
    public MULIU(int a, int b, int c) { super(Opcode.MULIU, a, b, c); }
}

class DIVIU extends OpRaRbUc {
    public DIVIU(int a, int b, int c) { super(Opcode.DIVIU, a, b, c); }
}

class MODIU extends OpRaRbUc {
    public MODIU(int a, int b, int c) { super(Opcode.MODIU, a, b, c); }
}

class CMPIU extends OpRaRbUc {
    public CMPIU(int a, int b, int c) { super(Opcode.CMPIU, a, b, c); }
}

class ORIU extends OpRaRbUc {
    public ORIU(int a, int b, int c) { super(Opcode.ORIU, a, b, c); }
}

class ANDIU extends OpRaRbUc {
    public ANDIU(int a, int b, int c) { super(Opcode.ANDIU, a, b, c); }
}

class BICIU extends OpRaRbUc {
    public BICIU(int a, int b, int c) { super(Opcode.BICIU, a, b, c); }
}

class XORIU extends OpRaRbUc {
    public XORIU(int a, int b, int c) { super(Opcode.XORIU, a, b, c); }
}

/**
 * <b><code>UnOpIU a uc</code></b><br>
 *    <code>R.a = UnOp uc</code>
 */
abstract class OpRa__xc extends RealInstruction {
    public OpRa__xc(Opcode opcode, int a, int c) {
        super(opcode, a, Integer.MIN_VALUE, c);
    }
}

class OpRa__Uc extends OpRa__xc {
    public OpRa__Uc(Opcode opcode, int a, int c) {
        super(opcode, a, c);
    }
    protected String argsToString() {
        return regToString(a) + ", " + (c & 0xFFFF);
    }
}

class CHKIU extends OpRa__Uc {
    public CHKIU(int a, int c) { super(Opcode.CHKIU, a, c); }
}

//########################################################################
// Opcodes - load/store instructions

class LDW extends OpRaRbIc {
    public LDW(int a, int b, int c) { super(Opcode.LDW, a, b, c); }
}

class LDB extends OpRaRbIc {
    public LDB(int a, int b, int c) { super(Opcode.LDB, a, b, c); }
}

class POP extends OpRaRbIc {
    public POP(int a, int b, int c) { super(Opcode.POP, a, b, c); }
}

class STW extends OpRaRbIc {
    public STW(int a, int b, int c) { super(Opcode.STW, a, b, c); }
}

class STB extends OpRaRbIc {
    public STB(int a, int b, int c) { super(Opcode.STB, a, b, c); }
}

class PSH extends OpRaRbIc {
    public PSH(int a, int b, int c) { super(Opcode.PSH, a, b, c); }
}

//########################################################################
/**
 * <b><code>Branch a oc</code></b><br>
 *    <code>If bop a: PC = PC + 4*oc</code>
 */
abstract class OpRa__Oc extends RealInstruction {
    protected final String label;

    public OpRa__Oc(Opcode opcode, int a, String label) {
        super(opcode, a, Integer.MIN_VALUE, Integer.MIN_VALUE);
        this.label = label;
    }
    protected String argsToString() {
        return regToString(a) + ", " + label;
    }
}

//########################################################################
// Opcodes - control instructions

class BEQ extends OpRa__Oc {
    public BEQ(int a, String c) { super(Opcode.BEQ, a, c); }
}

class BNE extends OpRa__Oc {
    public BNE(int a, String c) { super(Opcode.BNE, a, c); }
}

class BLT extends OpRa__Oc {
    public BLT(int a, String c) { super(Opcode.BLT, a, c); }
}

class BGE extends OpRa__Oc {
    public BGE(int a, String c) { super(Opcode.BGE, a, c); }
}

class BLE extends OpRa__Oc {
    public BLE(int a, String c) { super(Opcode.BLE, a, c); }
}

class BGT extends OpRa__Oc {
    public BGT(int a, String c) { super(Opcode.BGT, a, c); }
}

abstract class Op____Oc extends RealInstruction {
    protected final String label;

    public Op____Oc(Opcode opcode, String label) {
        super(opcode, Integer.MIN_VALUE, Integer.MIN_VALUE, Integer.MIN_VALUE);
        this.label = label;
    }
    protected String argsToString() {
        return label;
    }
}

class BSR extends Op____Oc {
    public BSR(String c) { super(Opcode.BSR, c); }
}

abstract class Op____Lc extends RealInstruction {
    protected final String label;

    public Op____Lc(Opcode opcode, String label) {
        super(opcode, Integer.MIN_VALUE, Integer.MIN_VALUE, Integer.MIN_VALUE);
        this.label = label;
    }
    protected String argsToString() {
        return label;
    }
}

class JSR extends Op____Lc {
    public JSR(String c) { super(Opcode.JSR, c); }
}

abstract class Op____Rc extends RealInstruction {
    public Op____Rc(Opcode opcode, int c) {
        super(opcode, Integer.MIN_VALUE, Integer.MIN_VALUE, c);
    }
    protected String argsToString() {
        return regToString(c);
    }
}

class RET extends Op____Rc {
    public RET(int c) { super(Opcode.RET, c); }
    public BitSet uses() {
        // Claim that RET uses the result register, to make sure it's
        // not considered dead.
        BitSet uses = super.uses();
        uses.set(RISC.RES);
        return uses;
    }
}

//########################################################################
// Opcodes - miscellaneous instructions

/**
 * <p>Opcode: stop execution and return to debugger.</p>
 * <b><code>BREAK</code></b><br>
 *    stop execution and return to debugger
 */
class BREAK extends RealInstruction {
    public BREAK() {
        super(Opcode.BREAK,
              Integer.MIN_VALUE,
              Integer.MIN_VALUE,
              Integer.MIN_VALUE);
    }
    protected String argsToString() { return ""; }
}

/**
 * <p>Opcode: Name a place in the code..</p>
 * <b><code>LABEL n</code></b><br>
 */

class LABEL extends RealInstruction {
    protected final String label;

    private static final Opcode MyOpcode =
        new Opcode(-1, -1, "LABEL", Opcode.REG_NONE, Opcode.REG_NONE, 1);

    public LABEL(String label) {
        super(MyOpcode,
              Integer.MIN_VALUE,
              Integer.MIN_VALUE,
              Integer.MIN_VALUE);
        this.label = label;
    }

    public String toString() {
        return label + ":";
    }
    protected String argsToString() { throw new Error(); }
}


/**
 * <p>Opcode: invoke a system function.</p>
 * <b><code>SYSCALL a b uc</b></code><br>
 *    invoke system function <code>uc</code> with registers
 *    <code>R.a</code> and <code>R.b</code>
 */

class SYSCALL extends OpRaRbUc {
    public SYSCALL(int a, int b, int c) {
        super(Opcode.SYSCALL, a, b, c);
    }
    public BitSet defs() {
        switch(c) {
        case RISC.SYS_IO_RD_CHR:
        case RISC.SYS_IO_RD_INT:
        case RISC.SYS_IO_RD_BIN:
        case RISC.SYS_IO_RD_DEC:
        case RISC.SYS_IO_RD_HEX:
        case RISC.SYS_GC_ALLOC:
        case RISC.SYS_GET_TOTAL_MEM_SIZE:
            BitSet bs = new BitSet();
            bs.set(a);
            return bs;

        case RISC.SYS_IO_WR_CHR:
        case RISC.SYS_IO_WR_INT:
        case RISC.SYS_IO_WR_BIN:
        case RISC.SYS_IO_WR_DEC:
        case RISC.SYS_IO_WR_HEX:
        case RISC.SYS_IO_WR_LN:
        case RISC.SYS_IO_FLUSH:
        case RISC.SYS_GC_INIT:
        case RISC.SYS_EXIT:
        case RISC.SYS_DB_WR_REG:
        case RISC.SYS_DB_DUMP:
            return new BitSet();
        default:
            throw new Error();
        }
    }
    public BitSet uses() {
        BitSet bs = new BitSet();
        switch(c) {
        case RISC.SYS_IO_RD_CHR:
        case RISC.SYS_IO_RD_INT:
        case RISC.SYS_IO_RD_BIN:
        case RISC.SYS_IO_RD_DEC:
        case RISC.SYS_IO_RD_HEX:
        case RISC.SYS_IO_WR_LN:
        case RISC.SYS_IO_FLUSH:
        case RISC.SYS_GET_TOTAL_MEM_SIZE:
        case RISC.SYS_DB_WR_REG:
        case RISC.SYS_DB_DUMP:
            break;
        case RISC.SYS_EXIT:
        case RISC.SYS_IO_WR_CHR:
        case RISC.SYS_IO_WR_INT:
        case RISC.SYS_IO_WR_BIN:
        case RISC.SYS_IO_WR_DEC:
        case RISC.SYS_IO_WR_HEX:
            bs.set(a);
            break;
        case RISC.SYS_GC_INIT:
            bs.set(a); /* !! Fallthrough !! */
        case RISC.SYS_GC_ALLOC:
            bs.set(b);
            break;
        }
        return bs;
    }
    public void replaceRegs(int[] map) {
        if (a >= 0 && a < map.length) a = map[a];
        if (b >= 0 && b < map.length) b = map[b];
    }
}

//########################################################################
// !!! Pseudo instructions

abstract class PseudoInstruction extends Instruction {
    public static final Opcode LDA =
        new Opcode(-1, RISC.LDA, "lda", Opcode.REG_A, Opcode.REG_NONE, 1);
    public static final Opcode CALL =
        new Opcode(-1, RISC.CALL, "call", Opcode.REG_NONE, Opcode.REG_C, 1, true);
    public static final Opcode PCALL =
        new Opcode(-1, -1, "pcall", Opcode.REG_A, Opcode.REG_C, 1, true);
    public static final Opcode PJSR =
        new Opcode(-1, -1, "pjsr", Opcode.REG_A, Opcode.REG_C, 1, true);

    public PseudoInstruction(Opcode opcode, int a, int b, int c) {
        super(opcode, a, b, c);
    }
}

abstract class OpRa__Lc extends PseudoInstruction {
    protected final String label;

    public OpRa__Lc(Opcode opcode, int a, String label) {
        super(opcode, a, Integer.MIN_VALUE, Integer.MIN_VALUE);
        this.label = label;
    }
    protected String argsToString() {
        return regToString(a) + ", " + label;
    }
}

class LDA extends OpRa__Lc {
    public LDA(int a, String c) {
        super(PseudoInstruction.LDA, a, c);
    }
    public int size() { return 1; }
}

class CALL extends Op____Rc {
    public CALL(int c) {
        super(PseudoInstruction.CALL, c);
    }
    public int size() { return 2; }
}

class PseudoCall extends PseudoInstruction {
    public int resReg;
    public List<Integer> args;

    public PseudoCall(Opcode opcode, int resReg, List<Integer> args) {
        super(opcode, Integer.MIN_VALUE, Integer.MIN_VALUE, Integer.MIN_VALUE);
        this.resReg = resReg;
        this.args = args;
    }
    public int size() {
        return 2 + args.size();
    }
    public BitSet defs() {
        BitSet bs = new BitSet();
        bs.set(RISC.RES);
        bs.set(resReg);
        return bs;
    }
    public BitSet uses() {
        BitSet bs = new BitSet();
        Iterator<Integer> argsIt = args.iterator();
        while (argsIt.hasNext()) {
            Integer arg = (Integer) argsIt.next();
            bs.set(arg.intValue());
        }
        return bs;
    }
    public void replaceRegs(int[] map) {
        if (resReg < map.length) resReg = map[resReg];
        Iterator<Integer> argsIt = args.iterator();
        args = new LinkedList<Integer>();
        while (argsIt.hasNext()) {
            int r = ((Integer)argsIt.next()).intValue();
            int newR = (r < map.length) ? map[r] : r;
            args.add(new Integer(newR));
        }
    }
    protected String argsToString() {
        StringBuffer buf = new StringBuffer(regToString(resReg));
        Iterator<Integer> argsIt = args.iterator();
        while (argsIt.hasNext()) {
            buf.append(", ");
            buf.append(regToString((/*(Integer)*/argsIt.next()).intValue()));
        }
        return buf.toString();
    }
}

class PCALL extends PseudoCall {
    public int addrReg;

    public PCALL(int res, List<Integer> args, int addrReg) {
        super(PseudoInstruction.PCALL, res, args);
        this.addrReg = addrReg;
    }
    public BitSet uses() {
        BitSet bs = super.uses();
        bs.set(addrReg);
        return bs;
    }
    public void replaceRegs(int[] map) {
        super.replaceRegs(map);
        addrReg = map[addrReg];
    }
    protected String argsToString() {
        String s = super.argsToString();
        return s + (s.equals("") ? "" : ", ") + regToString(addrReg);
    }
}

class PJSR extends PseudoCall {
    protected final String label;

    public PJSR(int res, List<Integer> args, String label) {
        super(PseudoInstruction.PJSR, res, args);
        this.label = label;
    }
    protected String argsToString() {
        String s = super.argsToString();
        return s + (s.equals("") ? "" : ", ") + label;
    }
}
