package einsc.backend.cfg;

import java.util.LinkedList;
import java.util.Iterator;
import einsc.*;
import einsc.backend.*;

/**
 * Linearise control-flow graphs.
 */
public class Lineariser implements RISC {

    /**
     * Linearise a complete program made up of the given control-flow
     * graphs. Code to initialise the garbage collector and stack
     * pointer is output first, followed by the predefined functions,
     * and finally the linearised cfgs.
     *
     * @param globalScope The program's global scope, containing the
     * predefined functions (and nothing else).
     * @param cfgs The control-flow graphs which together make the
     * program. One of them must be named _Main, and its code will be
     * run first.
     * @return The linearised code.
     */
    public Code lineariseProgram(Scope globalScope, CFG[] cfgs) {
        Code code = new Code();

        // Output initialization code

        // initialise stack pointer
        code.emit(SYSCALL, SP, 0, SYS_GET_TOTAL_MEM_SIZE);
        // initialise garbage collector
        code.emit(DIVI, 1, SP, 4);
        code.emit(DIVI, 2, 1, 2);
        code.emit(ADDI, 3, 0, SP);
        code.emit(LSHI, 3, 3, 27);
        code.emit(ADD, 2, 2, 3);
        code.emit(SYSCALL, 1, 2, SYS_GC_INIT);

        // jump to main expression
        code.emit(BEQ, ZERO, "_Main");

        // Output predefined functions
        for (Symbol sym = globalScope.first(); sym != null; sym = sym.next) {
        	code.emit(LABEL,sym.label);

            if (sym.name.equals("printInt")) {
                code.emit(POP, 1, SP, WORD_SIZE);
                code.emit(SYSCALL, 1, ZERO, SYS_IO_WR_INT);
                code.emit(ADD, RES, ZERO, ZERO);
            } else if (sym.name.equals("printChar")) {
                code.emit(POP, 1, SP, WORD_SIZE);
                code.emit(SYSCALL, 1, ZERO, SYS_IO_WR_CHR);
                code.emit(ADD, RES, ZERO, ZERO);
            } else if (sym.name.equals("readInt")) {
                code.emit(SYSCALL, RES, ZERO, SYS_IO_RD_INT);
            } else if (sym.name.equals("readChar")) {
                code.emit(SYSCALL, RES, ZERO, SYS_IO_RD_CHR);
            } else if (sym.kind == Symbol.FUNC)
                throw new Error("unknown predefined function: " + sym.name);
            code.emit(RET, LNK);
        }

        // Output user functions and main expression
        for (int i = 0; i < cfgs.length; ++i)
            lineariseCFG(cfgs[i], code);

        return code;
    }

    /**
     * Linearise the given cfg and append the result to the given
     * code. Linearisation is done by outputting the nodes in reverse
     * pre-order.
     *
     * @param cfg The CFG to linearise.
     * @param code The code to which to append the result.
     */
    public void lineariseCFG(CFG cfg, Code code) {
        CFG.Node[] nodes = reversePostOrderNodes(cfg);
		code.emit(LABEL, cfg.name);
		for (int i = 0; i < nodes.length; ++i) {
			CFG.Node node = nodes[i];
			code.emit(LABEL, cfg.name + "_" + node.label);
			code.append(node.blockIterator());

			Iterator<CFG.Edge> edgeIt = cfg.getEdgesFrom(node).iterator();
			while (edgeIt.hasNext()) {
				CFG.Edge edge = /*(CFG.Edge) */edgeIt.next();
				if ((i + 1) == nodes.length || edge.to != nodes[i + 1])
					code.emit(edge.branch);
			}
		}
    }

    private CFG.Node[] reversePostOrderNodes(CFG cfg) {
        final LinkedList<CFG.Node> nodes = new LinkedList<CFG.Node>();
        cfg.visitDFS(new CFG.EmptyVisitor() {
                public void post(CFG.Node node) { nodes.addFirst(node); }
            });
        return (CFG.Node[])nodes.toArray(new CFG.Node[nodes.size()]);
    }
}
