package einsc.backend.opt;

import java.util.*;
import einsc.backend.cfg.*;

abstract public class DataFlowAnalysis {
    protected final CFG cfg;

    private final HashMap<CFG.Node, Object> in = new HashMap<CFG.Node, Object>();
    private final HashMap<CFG.Node, Object> out = new HashMap<CFG.Node, Object>();

    public DataFlowAnalysis(CFG cfg) {
        this.cfg = cfg;
    }

    public Object in(CFG.Node n) {
        return in.get(n);
    }

    public Object out(CFG.Node n) {
        return out.get(n);
    }

    public void analyseBackwards() {
        final HashMap<CFG.Node, Transfer> transferFunctions =
            new HashMap<CFG.Node, Transfer>();

        /* Initialise the analysis.
           Calculate transfer functions for each BB.
           Store the result in transferFunctions,
           a map Node->Transfer. */
        cfg.visitDFS(new CFG.EmptyVisitor() {
                public void pre(CFG.Node node) {
                    /* Calculate transfer functions for this node. */
                    transferFunctions.put(node, calculateTransfer(node));
                    /* Initialise in and out.*/
                    in.put(node, newValue());
                    out.put(node, newValue());
                }
            });

        /* Fix-point iteration **/
        LinkedList<CFG.Node> worklist = new LinkedList<CFG.Node>(cfg.getNodes());
        while (! worklist.isEmpty()) {
            /* Take an element (n) from the worklist. */
            CFG.Node n = (CFG.Node) worklist.removeFirst();
            Object oldIn = cloneValue(in(n));
            Object newOut = out(n);

            /* Apply convergence function: U(succ(n).in) */
            Iterator<CFG.Node> succIt = cfg.successors(n).iterator();
            while (succIt.hasNext())
                newOut = mergeValues(newOut, in(/*(CFG.Node)*/succIt.next()));
            out.put(n, newOut);

            /* Apply transfer function */
            Object newIn =
                (((Transfer)transferFunctions.get(n)).apply(newOut));
            in.put(n, newIn);

            /* If "in" changed, add predecessors to worklist. */
            if (! newIn.equals(oldIn)) {
                List<CFG.Node> preds = cfg.predecessors(n);
                preds.removeAll(worklist);
                worklist.addAll(preds);
            }
        }
    }

    /** Create a new analysis value */
    abstract protected Object newValue();
    /** Clone an existing analysis value */
    abstract protected Object cloneValue(Object v);
    /** Merge two analysis values */
    abstract protected Object mergeValues(Object v1, Object v2);

    abstract protected Transfer calculateTransfer(CFG.Node node);

    protected abstract class Transfer {
        abstract Object apply(Object input);
        public String toString() { return "Object -> Object"; }
    }
}
